from xmlrpc.client import boolean
import torch
from random import shuffle
import numpy as np

## --- Default Architectures ---
# `**hp` must be included

# `select_fn_lose()`
def lose_binary(**hp):
    loser = torch.nn.BCELoss()
    return loser

def lose_multiclass(**hp):
    # ptrckblck says `nn.NLLLoss()` will work too.
    loser = torch.nn.CrossEntropyLoss()
    return loser

def lose_regression(**hp):
    loser = torch.nn.L1Loss()#mean absolute error.
    return loser

# `select_fn_optimize()`
def optimize(model, **hp):
    optimizer = torch.optim.Adamax(model.parameters(),lr=0.01)
    return optimizer

# `select_fn_predict()`
def predict_binary(model, features):
    probability = model(features)
    # Convert tensor back to numpy for AIQC metrics.
    probability = probability.detach().numpy()
    prediction  = (probability > 0.5).astype("int32")
    # Both objects are numpy.
    return prediction, probability

def predict_multiclass(model, features):
    probabilities = model(features)
    # Convert tensor back to numpy for AIQC metrics.
    probabilities = probabilities.detach().numpy()
    prediction    = np.argmax(probabilities, axis=-1)
    # Both objects are numpy.
    return prediction, probabilities

def predict_regression(model, features):
    prediction = model(features).detach().numpy()
    return prediction


## --- Batching and Training Loop ---
def drop_invalidBatch(
    batched_data:object
    , batch_size:int
    , enforce_sameSize:bool   = True
    , allow_singleSample:bool = False
):
    """
    `enforce_sameSize=True` Primarily because it influence batch size, therefore layer dimensions, 
    and also because true tensors must have uniform shape.
    """
    # Similar to a % remainder, this will only apply to the last element in the batch.
    last_batch_size = batched_data[-1].shape[0]
    # If there is a problem, then just trim the last split.
    if (last_batch_size==1):
        if (allow_singleSample==True):
            print("\nWarning - The size of the last batch is 1,\n which commonly leads to PyTorch errors.\nTry using `torch_batch_samples(allow_singleSample=False)\n")
        elif (allow_singleSample==False): 
            batched_data = batched_data[:-1]
    elif ((enforce_sameSize==True) and (batch_size!=last_batch_size)):
        batched_data = batched_data[:-1]
    return batched_data


def shuffle_samples(features:list, label:object):
    """Assumes that the first index represents the batch."""
    rand_idx = list(range(len(label)))
    shuffle(rand_idx)
    
    label = torch.stack([label[i] for i in rand_idx]).to(torch.float)

    if isinstance(features, list):
        features = []
        for f in features:
            f = torch.stack([f[i] for i in rand_idx]).to(torch.float)
            features.append(f)
    else:
        torch.stack([features[i] for i in rand_idx]).to(torch.float)
    return features, label


def batch_samples(
    features:object
    , label:object 
    , batch_size              = 5
    , enforce_sameSize:bool   = False
    , allow_singleSample:bool = False
):
    if (batch_size==1):
        if (allow_singleSample==False):
            msg = "\nYikes - `batch_size==1` but `allow_singleSample==False`.\n"
            raise Exception(msg)
        elif (allow_singleSample==True):
            msg = "\nWarning - PyTorch errors are common when `batch_size==1`.\n"
            print(msg)
    
    # split() normally returns a tuple.
    label = list(torch.split(label, batch_size))
    label = drop_invalidBatch(label, batch_size, enforce_sameSize, allow_singleSample)

    if isinstance(features, list):
        batched_features = []
        for f in features:
            f = list(torch.split(f, batch_size))
            f = drop_invalidBatch(f, batch_size, enforce_sameSize, allow_singleSample)
            batched_features.append(f)
        features = []
        for bf in batched_features:
            for f in bf:
                features.append(f)
    else:
        features = list(torch.split(features, batch_size))
        features = drop_invalidBatch(features, batch_size, enforce_sameSize, allow_singleSample)
    return features, label


def shuffle_batches(features:list, label:list):
    """
    - Assumes that the first index represents the batch.
    - Makes sure batches aren't seen in same order every epoch.
    """
    rand_idx = list(range(len(label)))
    shuffle(rand_idx)
    
    label = [label[i] for i in rand_idx]
    # This time multiple features don't need special treatment.
    features = [features[i] for i in rand_idx]
    return features, label


def flatten_uniColumn(tzr:object):
    if (tzr.shape[-1]==1):
        return tzr.flatten()
    return tzr


def flip_floatInt(tzr:object):
    """
    Handles float/int incosistencies of torch's loss and torchmetrics' scoring.
    For example binary classify (model/loser=float) but (torchmetric_accuracy=int)
    github.com/PyTorchLightning/metrics/discussions/1059
    """
    if (tzr.type()=='torch.FloatTensor'):
        # Sample to see if floats are ints decimals e.g. `2.`
        # `float(0).is_integer()==True`
        if all([float(i).is_integer() for i in tzr[:3]]):
            return tzr.to(torch.int64)
        else:
            msg = f"\nYikes - Scoring failed on {tzr.type()}.\nDid not attempt as int64 because tensor contained non-zero decimals.\n"
            raise Exception(msg)
    elif (tzr.type()=='torch.LongTensor'):
        return tzr.to(torch.float32)

    else:
        msg = f"\nYikes - Scoring failed because {tzr.type()} type not supported.\n"
        raise Exception(msg)


def fit(
    model:object
    , loser:object
    , optimizer:object
    , train_features:list
    , train_label:object
    , eval_features:list
    , eval_label:object
    , epochs:int              = 30
    , batch_size:int          = 5
    , enforce_sameSize:bool   = True
    , allow_singleSample:bool = False
    , metrics:list            = None
):
    """
    - This is the only user-facing function for non-manual training loops.
    - It is designed to handle all supervised scenarios.
    - Have not tested this with self-supervised where 2D+ compared to 2D+
    - Need to write a test for multi-modal where `train_features` is a list of tensors.
    """
    # Mirrors `tf.keras.model.History.history` schema for use with `Predictor.plot_learning_curve()`
    history = dict(loss=list(), val_loss=list())
    metrics_keys = []
    if (metrics is not None):
        for m in metrics:
            # An initialized metric actually contains `None` so `utils.listify` doesn't work here.
            if ('torchmetrics' not in str(type(m))):
                msg = "\nYikes - Did you forget to initialize your metric?\ne.g. do `torchmetrics.Accuracy()`, not `torchmetrics.Accuracy`\n"
                raise Exception(msg)
            name              = m.__class__.__name__
            history[name]     = list()
            val_name          = f"val_{name}"
            history[val_name] = list()
            metrics_keys.append((name, val_name))

    train_features, train_label = shuffle_samples(train_features, train_label)
    eval_features, eval_label = shuffle_samples(eval_features, eval_label)
    ## --- Prepare mini batches for analysis ---
    # The variables below are reassigned because we need the above variables for epoch metrics
    trainFeatures_batched, trainLabel_batched = batch_samples(
        train_features, train_label,
        batch_size=batch_size, enforce_sameSize=enforce_sameSize, allow_singleSample=allow_singleSample
    )
    evalFeatures_batched, evalLabel_batched = batch_samples(
        eval_features, eval_label,
        batch_size=batch_size, enforce_sameSize=enforce_sameSize, allow_singleSample=allow_singleSample
    )
    """
    - On one hand, I could pass in `analysis_type` to deterministically handle the proper
      dimensionality and type of the data for loss and metrics.
    - However, performance-wise, that would still require a lot of if statements. 
    - The `try` approach is more future-proof.
    - `flatten()` succeeds on any dimension, even 1D.
    - Remember, multi-label  PyTorch uses ordinal labels, but OHE output probabilities.
      It wants to compare 2D probabilities to 1D ordinal labels.
    - Unsupervised analysis either succeeds as 2D+ or fails. MSE works on 3D data, but r2 fails. 
      We could stack 3D+ into 2D, but then we'd have to stack features as well, and that's kind 
      of insane because this is just for epoch-level metrics.
    """
    ## --- Training Loop ---
    for epoch in range(epochs):
        trainFeatures_batched, trainLabel_batched = shuffle_batches(trainFeatures_batched, trainLabel_batched)
        evalFeatures_batched, evalLabel_batched = shuffle_batches(evalFeatures_batched, evalLabel_batched)

        ## --- Batch Training ---
        for e, batch_features in enumerate(trainFeatures_batched):
            # Make raw (unlabeled) predictions.
            batch_label       = trainLabel_batched[e]
            batch_probability = model(batch_features)
            try:
                batch_loss = loser(batch_probability, batch_label)
            except:
                # Known exception: multi classify fails on 2D and floats
                batch_label = flatten_uniColumn(batch_label)
                batch_label = flip_floatInt(batch_label)
                batch_loss  = loser(batch_probability, batch_label)
            # Backpropagation.
            optimizer.zero_grad()
            batch_loss.backward()
            optimizer.step()
        
        ## --- Epoch Loss ---
        # These need not be shuffled/ batched during each epoch
        train_probability = model(train_features)
        train_probability = flatten_uniColumn(train_probability)
        train_label       = flatten_uniColumn(train_label)
        try:
            train_loss = loser(train_probability, train_label)
        except:
            # Known exception: multi classify fails on float
            # Known exception: binary classify fails on int. inconsistent w torchmetrics accuracy.
            train_label = flip_floatInt(train_label)
            train_loss  = loser(train_probability, train_label)
        history['loss'].append(float(train_loss))

        eval_probability = model(eval_features)
        eval_probability = flatten_uniColumn(eval_probability)
        eval_label       = flatten_uniColumn(eval_label)
        try:
            eval_loss = loser(eval_probability, eval_label)
        except:
            eval_label = flip_floatInt(eval_label)
            eval_loss  = loser(eval_probability, eval_label)
        history['val_loss'].append(float(eval_loss))

        ## --- Epoch Metrics ---
        # Known exception: binary classify accuracy fails on float. inconsistent w model/loser.
        for i, m in enumerate(metrics):
            try:
                train_m = m(train_probability, train_label)
            except:
                train_label = flip_floatInt(train_label)
                train_m     = m(train_probability, train_label)
            metrics_key = metrics_keys[i][0]
            history[metrics_key].append(float(train_m))

            try:
                eval_m = m(eval_probability, eval_label)
            except:
                eval_label = flip_floatInt(eval_label)
                eval_m     = m(eval_probability, eval_label)
            metrics_key = metrics_keys[i][1]
            history[metrics_key].append(float(eval_m))
    return model, history



"""
Serialization
└── Documentation = https://dill.readthedocs.io/en/latest/index.html

Goes beyond pickle to handle nested functions and classes.
Dill-serialized objects are pickleable.
"""
from io import BytesIO
from dill import dump, load, source


def serialize(objekt:object):
    blob = BytesIO()
    dump(objekt, blob)
    blob = blob.getvalue()
    return blob


def deserialize(blob:bytes):
    objekt = BytesIO(blob)
    objekt = load(objekt)
    return objekt


def reveal_code(blob:object):
    code_str = source.getsource(
        deserialize(blob).__code__
    )
    return code_str



"""
Datum
└── Documentation = https://aiqc.readthedocs.io/en/latest/notebooks/datasets.html

- Example datasets that are used for both tutorials and tests.
- Either included in the aiqc package `aiqc/data` or stored remotely `/remote_datum`
"""
#`importlib.resources` was not working on Google Collab.
from pkg_resources import resource_filename 
import pandas as pd


def list_datums(format:str=None):
    """
    - Of course, these could be classes, but record format let's us display all of the info in a dataframe.
    - 'name' value cannot include 'https' because that's how remote datasets are detected.
    """
    records = [
        dict(
            name = 'exoplanets.parquet'
            , dataset_type  = 'tabular'
            , analysis_type = 'regression'
            , label         = 'SurfaceTempK'
            , label_classes = 'N/A'
            , features      = 8
            , samples       = 433
            , description   = 'Predict temperature of exoplanet.'
            , location      = 'local'
        ),
        dict(
            name            = 'heart_failure.parquet'
            , dataset_type  = 'tabular'
            , analysis_type = 'regression'
            , label         = 'died'
            , label_classes = '2'
            , features      = 12
            , samples       = 299
            , description   = "Biometrics to predict loss of life."
            , location      = 'local'
        ),
        dict(
            name            = 'iris.tsv'
            , dataset_type  = 'tabular'
            , analysis_type = 'classification_multi'
            , label         = 'species'
            , label_classes = 3
            , features      = 4
            , samples       = 150
            , description   = '3 species of flowers. Only 150 rows, so cross-folds not represent population.'
            , location      = 'local'
        ),
        dict(
            name            = 'sonar.csv'
            , dataset_type  = 'tabular'
            , analysis_type = 'classification_binary'
            , label         = 'object'
            , label_classes = 2
            , features      = 60
            , samples       = 208
            , description   = 'Detecting either a rock "R" or mine "M". Each feature is a sensor reading.'
            , location      = 'local'
        ),
        dict(
            name            = 'houses.csv'
            , dataset_type  = 'tabular'
            , analysis_type = 'regression'
            , label         = 'price'
            , label_classes = 'N/A'
            , features      = 12
            , samples       = 506
            , description   = 'Predict the price of the house.'
            , location      = 'local'
        ),
        dict(
            name            = 'iris_noHeaders.csv' 
            , dataset_type  = 'tabular'
            , analysis_type = 'classification multi'
            , label         = 'species'
            , label_classes = 3
            , features      = 4
            , samples       = 150
            , description   = 'For testing; no column names.'
            , location      = 'local'
        ),
        dict(
            name            = 'iris_10x.tsv'
            , dataset_type  = 'tabular'
            , analysis_type = 'classification multi'
            , label         = 'species'
            , label_classes = 3
            , features      = 4
            , samples       = 1500
            , description   = 'For testing; duplicated 10x so cross-folds represent population.'
            , location      = 'local'
        ),
        dict(
            name            = 'brain_tumor.csv'
            , dataset_type  = 'image'
            , analysis_type = 'classification_binary'
            , label         = 'status'
            , label_classes = 2
            , features      = '1 color x 160 tall x 120 wide'
            , samples       = 80
            , description   = 'csv acts as label and manifest of image urls.'
            , location      = 'remote'
        ),
        dict(
            name            = 'galaxies.tsv'
            , dataset_type  = 'image'
            , analysis_type = 'classification_binary'
            , label         = 'morphology'
            , label_classes = 2
            , features      = '3 colors x 240 tall x 300 wide'
            , samples       = 40
            , description   = 'tsv acts as label and manifest of image urls.'
            , location      = 'remote'
        ),
        dict(
            name            = 'spam.csv'
            , dataset_type  = 'text'
            , analysis_type = 'classification_binary'
            , label         = 'v1'
            , label_classes = 2
            , features      = 'text data'
            , samples       = 5572
            , description   = 'collection of spam/ ham (not spam) messages'
            , location      = 'local'
        ),
        dict(
            name            = 'epilepsy.parquet'
            , dataset_type  = 'sequence'
            , analysis_type = 'classification_binary'
            , label         = 'seizure'
            , label_classes = 2
            , features      = '1 x 178 readings'
            , samples       = 1000
            , description   = "<https://archive.ics.uci.edu/ml/datasets/Epileptic+Seizure+Recognition> Storing the data tall so that it compresses better.`label_df = df[['seizure']]; sensor_arr3D = df.drop(columns=['seizure']).to_numpy().reshape(1000,178,1)`"
            , location      = 'local'
        ),
        dict(
            name            = 'delhi_climate.parquet'
            , dataset_type  = 'sequence'
            , analysis_type = 'forecasting'
            , label         = 'N/A'
            , label_classes = 'N/A'
            , features      = '3'
            , samples       = 1575
            , description   = "<https://www.kaggle.com/sumanthvrao/daily-climate-time-series-data>. Both train and test (pruned last day from train). 'pressure' and 'wind' columns seem to have outliers. Converted 'date' column to 'day_of_year.'"
            , location      = 'local'
        ),
        dict(
            name            = 'liberty_moon.csv'
            , dataset_type  = 'image'
            , analysis_type = 'forecasting'
            , label         = 'N/A'
            , label_classes = 'N/A'
            , features      = '1 color x 50 tall x 60 wide'
            , samples       = 15
            , description   = 'moon glides from top left to bottom right'
            , location      = 'remote'
        ),
    ]

    formats_df  = [None, 'pandas', 'df' ,'dataframe', 'd', 'pd']
    formats_lst = ['list', 'lst', 'l']
    if (format in formats_df):
        pd.set_option('display.max_column',100)
        pd.set_option('display.max_colwidth', 500)
        df = pd.DataFrame.from_records(records)
        return df
    elif (format in formats_lst):
        return records
    else:
        msg = f"\nYikes - The format you provided <{format}> is not one of the following:{formats_df} or {formats_lst}\n"
        raise Exception(msg)


def get_path(name:str):
    """
    - pkg_resources is used to dynamically find where the files are located on the system
    - In setup.py, `include_package_data=True,#triggers MANIFEST.in which grafts /data`
    - Remote datasets locations are just hardcoded for now.
    """
    if (name == 'brain_tumor.csv'):
        # 2nd aiqc is the repo, not the module.
        full_path = "https://raw.githubusercontent.com/aiqc/aiqc/main/remote_datum/image/brain_tumor/brain_tumor.csv"
    elif (name == 'galaxies.tsv'):
        full_path = "https://raw.githubusercontent.com/aiqc/aiqc/main/remote_datum/image/galaxy_morphology/galaxies.tsv"
    elif (name == 'liberty_moon.csv'):
        full_path = "https://raw.githubusercontent.com/aiqc/aiqc/main/remote_datum/image/liberty_moon/liberty_moon.csv"
    else:
        short_path = f"data/{name}"
        full_path = resource_filename('aiqc', short_path)
    return full_path


def to_df(name:str):
    file_path = get_path(name)

    if (('.tsv' in name) or ('.csv' in name)):
        if ('.tsv' in name):
            separator = '\t'
        elif ('.csv' in name):
            separator = ','
        else:
            separator = None
        df = pd.read_csv(file_path, sep=separator)
    elif ('.parquet' in name):
        df = pd.read_parquet(file_path)
    return df


def get_remote_urls(manifest_name:'str'):
    # `df=to_df` looks weird but it's right.
    df          = to_df(name=manifest_name)
    remote_urls = df['url'].to_list()
    return remote_urls



from ..mlops import Inference
# Register them all here so that they can be called like so: `from aiqc import tests`
from . import tf_bin_img, tf_bin_seq, tf_bin_tab, tf_fore_img, tf_fore_tab, tf_multi_tab, tf_reg_tab, torch_bin_img, \
    torch_bin_tab, torch_multi_tab, torch_reg_tab, torch_bin_seq


def infer(queue:int, include_label:bool=True):
    """Hack to run existing data back through pipeline for inference"""
    predictor = queue.jobs[0].predictors[0]
    splitset = queue.splitset

    if (include_label==True):
        if (splitset.label==None):
            raise Exception("\nYikes - you specified `include_label==True` but the orginal Pipeline had no Label.\n")
        label_dataset = splitset.label.dataset
    else:
        label_dataset = None
    feature_datasets = [f.dataset for f in splitset.features]

    prediction = Inference(predictor, feature_datasets, label_dataset)

    if (include_label==False):
        assert not isinstance(prediction.metrics, dict), \
            "\nYikes - Metrics aren't supposed to exist\n"
    elif (include_label==True):
        assert isinstance(prediction.metrics, dict), \
            "\nYikes - Metrics are supposed to exist\n"
    assert isinstance(prediction.predictions, dict), \
        "\nYikes - Predictions are supposed to exist\n"
    return prediction



"""TensorFlow Forecasting with Image data"""
# Internal modules
from ..mlops import Pipeline, Input, Stratifier, Experiment, Architecture, Trainer
from ..orm import Dataset
from ..utils.encoding import div255, mult255
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import FunctionTransformer


def fn_build(features_shape, label_shape, **hp):
	"""
	- Model size was too large for SQLite blob... need to rework to store it on the FS.
	- incoming features_shape = frame* channels * rows * columns
	- https://keras.io/api/layers/reshaping_layers/reshape/
	- Model size was too large for SQLite blob... need to rework to store it on the FS.
	- ConvLSTM1D is still in nightly build so use ConvLSTM2D
	- https://www.tensorflow.org/api_docs/python/tf/keras/layers/ConvLSTM2D
	- If data_format='channels_last' 5D tensor with shape: (samples, time, rows, cols, channels)
	"""
	m = tf.keras.models.Sequential()
	m.add(l.Conv1D(64*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))
	m.add(l.Conv1D(32*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))
	m.add(l.Conv1D(16*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))

	# decoding architecture
	m.add(l.Conv1D(16*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(32*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(64*hp['multiplier'], 3, activation=hp['activation']))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(50, 3, activation='relu', padding='same'))# removing sigmoid
	return m


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		optimizer = optimizer
		, loss    = loser
		, metrics = ['mean_squared_error']
	)
	
	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label)
		, verbose = 0
		, batch_size = hp['batch_size']
		, callbacks=[tf.keras.callbacks.History()]
		, epochs = hp['epoch_count']
	)
	return model



def make_queue(repeat_count:int=1, fold_count:int=None, permute_count=None):
	hyperparameters = dict(
		epoch_count  = [12]
		, batch_size = [3]
		, cnn_init   = ['he_normal']
		, activation = ['relu']
		, multiplier = [3]
	)
	
	
	folder_path = 'remote_datum/image/liberty_moon/images'
	
		# Just ensuring we test all forms of ingestion.
	if (fold_count is None):
		dataset = Dataset.Image.from_folder(
			folder_path=folder_path, ingest=False, retype='float64'
		)
	else:
		dataset = Dataset.Image.from_folder(
			folder_path=folder_path, ingest=True, retype='float64'
		)

	pipeline = Pipeline(
		inputs = Input(
			dataset  = dataset,
			window = Input.Window(size_window=1, size_shift=2),
			encoders = Input.Encoder(
				FunctionTransformer(div255, inverse_func=mult255),
				dtypes = 'float64'
			),
			reshape_indices = (0,3,4)
		),
		
		stratifier = Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)
	)

	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "regression"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline         = pipeline
			, repeat_count   = repeat_count
			, permute_count  = permute_count
			, search_percent = None
		)
	)
	return experiment



"""PyTorch Multi-label Classification with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..utils.pytorch import fit
from ..orm import Dataset
# External modules
import torch.nn as nn
import torchmetrics as tm
from sklearn.preprocessing import OrdinalEncoder, StandardScaler


def fn_build(features_shape, num_classes, **hp):
	model = nn.Sequential(
		nn.Linear(features_shape[-1], 12),
		nn.BatchNorm1d(12,12),
		nn.ReLU(),
		nn.Dropout(p=0.5),

		nn.Linear(12, num_classes),
		nn.Softmax(dim=1),
	)
	return model


def fn_lose(**hp):
	loser = nn.CrossEntropyLoss(reduction=hp['reduction'])
	return loser


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	return fit(
		model, loser, optimizer, 
		train_features, train_label,
		eval_features, eval_label,
		epochs=10, batch_size=hp['batch_size'],  
		metrics=[tm.Accuracy(), tm.F1Score()]
	)


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	hyperparameters = {
		"reduction":['mean'], "batch_size":[5]
	}
	
	if (fold_count is not None):
		file_path = datum.get_path('iris_10x.tsv')
	else:
		file_path = datum.get_path('iris.tsv')

	shared_dataset = Dataset.Tabular.from_path(
		file_path = file_path
	)

	pipeline = Pipeline(
		Input(
			dataset  = shared_dataset,
			encoders = Input.Encoder(
				StandardScaler(copy=False),
				dtypes = ['float64']
			)
		),

		Target(
			dataset = shared_dataset,
			column  = 'species',
			encoder = Target.Encoder(OrdinalEncoder())
		),

		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)
	)

	experiment = Experiment(
		Architecture(
			library           = "pytorch"
			, analysis_type   = "classification_multi"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment



"""Functions that help with dataset creation & validation"""
from .wrangle import listify
from os import path, listdir
from textwrap import dedent
from natsort import natsorted
from validators import url as val_url
from requests import get as requests_get
from PIL.Image import open as img_open
import numpy as np
import pandas as pd


def sorted_file_list(dir_path:str):
    if (not path.exists(dir_path)):
        msg = f"\nYikes - The path you provided does not exist according to `path.exists(dir_path)`:\n{dir_path}\n"
        raise Exception(msg)
    file_path = path.abspath(dir_path)
    if (path.isdir(file_path) == False):
        msg = f"\nYikes - The path that you provided is not a directory:{file_path}\n"
        raise Exception(msg)
    file_paths = listdir(file_path)
    # prune hidden files and directories.
    file_paths = [f for f in file_paths if not f.startswith('.')]
    file_paths = [f for f in file_paths if not path.isdir(f)]
    if not file_paths:
        msg = f"\nYikes - The directory that you provided has no files in it:{file_path}\n"
        raise Exception(msg)
    # folder path is already absolute
    file_paths = [path.join(file_path, f) for f in file_paths]
    file_paths = natsorted(file_paths)
    return file_paths


def arr_validate(ndarray:object):
    if (type(ndarray).__name__ != 'ndarray'):
        raise Exception("\nYikes - The `ndarray` you provided is not of the type 'ndarray'.\n")
    if (ndarray.dtype.names is not None):
        raise Exception(dedent("""
        Yikes - Sorry, we do not support NumPy Structured Arrays.
        However, you can use the `dtype` dict and `column_names` to handle each column specifically.
        """))
    if (ndarray.size==0):
        raise Exception("\nYikes - The ndarray you provided is empty: `ndarray.size == 0`.\n")


def colIndices_from_colNames(column_names:list, desired_cols:list):
    desired_cols = listify(desired_cols)
    col_indices = [column_names.index(c) for c in desired_cols]
    return col_indices


def df_stringifyCols(df:object, rename_columns:list=None):
    """
    - `rename_columns` would have been defined by the user during ingestion.
    - Pandas will assign a range of int-based columns if there are no column names.
      So I want to coerce them to strings because I don't want both string and int-based 
      column names for when calling columns programmatically, 
      and more importantly, 'Exception: parquet must have string column names'
    """
    cols_raw = df.columns.to_list()
    if (rename_columns is None):
        # in case the columns were a range of ints.
        cols_str = [str(c) for c in cols_raw]
    else:
        cols_str = [str(c) for c in rename_columns]
    # dict from 2 lists
    cols_dct = dict(zip(cols_raw, cols_str))
    
    df = df.rename(columns=cols_dct)
    columns = df.columns.to_list()
    return df, columns


def df_validate(dataframe:object, column_names:list):
    if (dataframe.empty):
        raise Exception("\nYikes - The dataframe you provided is empty according to `df.empty`\n")

    if (column_names is not None):
        col_count = len(column_names)
        structure_col_count = dataframe.shape[1]
        if (col_count != structure_col_count):
            raise Exception(dedent(f"""
            Yikes - The dataframe you provided has <{structure_col_count}> columns,
            but you provided <{col_count}> columns.
            """))


def df_setMetadata(dataframe:object, rename_columns:list=None, retype:object=None):
    shape = {}
    shape['samples'], shape['columns'] = dataframe.shape[0], dataframe.shape[1]
    """
    - Passes in user-defined columns in case they are specified.
    - Pandas auto-assigns int-based columns return a range when `df.columns`, 
      but this forces each column name to be its own str.
    """
    dataframe, columns = df_stringifyCols(df=dataframe, rename_columns=rename_columns)
    """
    - At this point, user-provided `dtype` can be either a dict or a singular string/ class.
    - The dtype dict does not need to address every column.
    - If columns are renamed, then the dtype dict must use the same renamed column names.
    - But a Pandas dataframe in-memory only has `dtypes` dict not a singular `dtype` str.
    - So we will ensure that there is 1 dtype per column.
    """
    if (retype is not None):
        # Accepts dict{'column_name':'dtype_str'}, string e.g. 'int64', or class `np.int64`.
        try:
            dataframe = dataframe.astype(retype)
        except BaseException as err:
            print(err)
            print("\nYikes - Failed to apply the dtypes you specified to the data you provided.\n")
            raise
        """
        Check if any user-provided dtype conversions failed in the actual dataframe dtypes
        Pandas dtype seems robust in comparing dtypes: 
        Even things like `'double' == dataframe['col_name'].dtype` will pass when `.dtype==np.float64`.
        Despite looking complex, category dtype converts to simple 'category' string.
        """
        if (not isinstance(retype, dict)):
            # Inspect each column:dtype pair and check to see if it is the same as the user-provided dtype.
            actual_dtypes = dataframe.dtypes.to_dict()
            for col_name, typ in actual_dtypes.items():
                if (typ != retype):
                    raise Exception(dedent(f"""
                    Yikes - You specified `dtype={retype},
                    but Pandas did not convert it: `dataframe['{col_name}'].dtype == {typ}`.
                    You can either use a different dtype, or try to set your dtypes prior to ingestion in Pandas.
                    """))
        elif (isinstance(retype, dict)):
            for col_name, typ in retype.items():
                if (typ != dataframe[col_name].dtype):
                    raise Exception(dedent(f"""
                    Yikes - You specified `dataframe['{col_name}']:dtype('{typ}'),
                    but Pandas did not convert it: `dataframe['{col_name}'].dtype == {dataframe[col_name].dtype}`.
                    You can either use a different dtype, or try to set your dtypes prior to ingestion in Pandas.
                    """))
    """
    Tested outlandish dtypes:
    - `DataFrame.to_parquet(engine='auto')` fails on:
        'complex', 'longfloat', 'float128'.
    - `DataFrame.to_parquet(engine='auto')` succeeds on:
        'string', np.uint8, np.double, 'bool'.
    
    - But the new 'string' dtype is not a numpy type!
        so operations like `np.issubdtype` and `StringArray.unique().tolist()` fail.
    """
    excluded_types = ['string', 'complex', 'longfloat', 'float128']
    actual_dtypes = dataframe.dtypes.to_dict().items()

    for col_name, typ in actual_dtypes:
        for et in excluded_types:
            if (et in str(typ)):
                raise Exception(dedent(f"""
                Yikes - You specified `dtype['{col_name}']:'{typ}',
                but aiqc does not support the following dtypes: {excluded_types}
                """))
    """
    Now, we take the all of the resulting dataframe dtypes and save them.
    Regardless of whether or not they were user-provided.
    Convert the classed `dtype('float64')` to a string so we can use it in `.to_df()`
    """
    dtype = {k: str(v) for k, v in actual_dtypes}
    
    # Each object has the potential to be transformed so each object must be returned.
    return dataframe, columns, shape, dtype


def path_to_df(
    file_path:str
    , file_format:str
    , header:object
):
    """
    - We do not know how the columns are named at source, so do not handle 
      the filtering of columns here.
    - Previously, I was using pyarrow for all tabular file formats. 
      However, it had worse support for missing column names and header skipping.
      So I switched to pandas for handling csv/tsv, but read_parquet()
      doesn't let you change column names easily, so using fastparquet for parquet.
    """ 
    if (not path.exists(file_path)):
        msg = f"\nYikes - The path you provided does not exist according to `path.exists(path)`:\n{file_path}\n"
        raise Exception(msg)
    if (not path.isfile(file_path)):
        msg = f"\nYikes - The path you provided is not a file according to `path.isfile(path)`:\n{file_path}\n"
        raise Exception(msg)

    if (file_format == 'tsv') or (file_format == 'csv'):
        if (file_format == 'tsv'):
            sep='\t'
        elif (file_format == 'csv'):
            sep=','
        df = pd.read_csv(
            filepath_or_buffer = file_path
            , sep              = sep
            , header           = header
        )

    elif (file_format == 'parquet'):
        if (header != 'infer'):
            raise Exception(dedent("""
            Yikes - The argument `header` is not supported for `file_format='parquet'`
            because Parquet stores column names as metadata.\n
            """))
        df = pd.read_parquet(path=file_path, engine='fastparquet')
    return df


def imgFolder_to_arr4D(folder:str):
    file_paths = sorted_file_list(folder)
    arr_3Ds = []
    formats = []
    for p in file_paths:
        img = img_open(p)
        formats.append(img.format)
        arr = np.array(img)
        # Coerce to 3D
        if (arr.ndim==2):
            arr=np.array([arr])
        arr_3Ds.append(arr)
    arr_4D = np.array(arr_3Ds)
    return arr_4D, formats


def imgURLs_to_arr4D(urls:list):
    arr_3Ds = []
    formats = []
    for url in urls:
        if (val_url(url) != True):
            msg = f"\nYikes - Invalid url detected within `urls` list:\n'{url}'\n"
            raise Exception(msg)
        raw_request = requests_get(url, stream=True).raw
        img = img_open(raw_request)
        formats.append(img.format)
        arr = np.array(img)
        # Coerce 3D
        if (arr.ndim==2):
            arr=np.array([arr])
        arr_3Ds.append(arr)
    arr_4D = np.array(arr_3Ds)
    return arr_4D, formats



"""TensorFlow Binary Classification with Image data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..utils.tensorflow import TrainingCallback
from ..orm import Dataset
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	# incoming features_shape = channels * rows * columns
	# https://keras.io/api/layers/reshaping_layers/reshape/
	# https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv1D
	# Conv1D shape = `batch_shape + (steps, input_dim)`
	m.add(l.Input(shape=features_shape))
	# drop channels
	m.add(l.Reshape(
		(features_shape[1],features_shape[2])
	))
	m.add(l.Conv1D(128*hp['neuron_multiply'], kernel_size=hp['kernel_size'], input_shape=features_shape, padding='same', activation='relu', kernel_initializer=hp['cnn_init']))
	m.add(l.MaxPooling1D(pool_size=hp['pool_size']))
	m.add(l.Dropout(hp['dropout']))
	
	m.add(l.Conv1D(256*hp['neuron_multiply'], kernel_size=hp['kernel_size'], padding='same', activation='relu', kernel_initializer=hp['cnn_init']))
	m.add(l.MaxPooling1D(pool_size=hp['pool_size']))
	m.add(l.Dropout(hp['dropout']))

	m.add(l.Flatten())
	m.add(l.Dense(hp['dense_neurons']*hp['neuron_multiply'], activation='relu'))
	m.add(l.Dropout(0.2))
	if hp['include_2nd_dense'] == True:
		m.add(l.Dense(hp['2nd_dense_neurons'], activation='relu'))

	m.add(l.Dense(units=label_shape[-1], activation='sigmoid'))
	return m


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		optimizer=optimizer
		, loss=loser
		, metrics=['accuracy']
	)

	metrics_cuttoffs = [
		{"metric":"val_accuracy", "cutoff":0.70, "above_or_below":"above"},
		{"metric":"accuracy", "cutoff":0.70, "above_or_below":"above"},
		{"metric":"val_loss", "cutoff":0.50, "above_or_below":"below"},
		{"metric":"loss", "cutoff":0.50, "above_or_below":"below"}
	]
	cutoffs = TrainingCallback.MetricCutoff(metrics_cuttoffs)
	
	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label)
		, verbose = 0
		, batch_size = hp['batch_size']
		, callbacks=[tf.keras.callbacks.History(), cutoffs]
		, epochs = hp['epoch_count']
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count=None):
	hyperparameters = {
		"include_2nd_dense": [True]
		, "neuron_multiply": [1.0]
		, "epoch_count": [10]
		, "learning_rate": [0.01]
		, "pool_size": [2]
		, "dropout": [0.4]
		, "batch_size": [8]
		, "kernel_size": [3]
		, "dense_neurons": [24]
		, "2nd_dense_neurons": [9]
		, "cnn_init": ['he_normal']
	}

	df = datum.to_df(name='brain_tumor.csv')
	label_dataset = Dataset.Tabular.from_df(dataframe=df)

	# Takes a while to run, but it tests both `from_urls` and `datum` functionality
	image_urls = datum.get_remote_urls(manifest_name='brain_tumor.csv')
	
	# Just ensuring we test all forms of ingestion.
	if (fold_count is None):
		feature_dataset = Dataset.Image.from_urls(urls=image_urls)
	else:
		feature_dataset = Dataset.Image.from_urls(urls=image_urls, ingest=False)
	
	pipeline = Pipeline(
		Input(
			dataset  = feature_dataset
		),
		
		Target(
			dataset = label_dataset,
			column = "status"
		),
		
		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)    
	)

	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline          = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment



from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..orm import Dataset
import torch.nn as nn
import torchmetrics as tm
from ..utils.pytorch import fit
from sklearn.preprocessing import StandardScaler


def fn_build(features_shape, label_shape, **hp):
	# LSTM() returns tuple of (tensor, (recurrent state))
	class extract_tensor(nn.Module):
		def forward(self,x):
			# Output shape (batch, features, hidden)
			tensor, _ = x
			# Reshape shape (batch, hidden)
			return tensor[:, -1, :]

	model = nn.Sequential(
		nn.LSTM(
			input_size = features_shape[-1],
			hidden_size = hp['hidden'],
			batch_first = True
		),
		extract_tensor(),
		nn.Linear(hp['hidden'], label_shape[-1]),
		nn.Sigmoid(),
	)
	return model


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):	return fit(
		model, loser, optimizer,
		train_features, train_label,
		eval_features, eval_label,
		epochs=hp['epochs'], batch_size=hp['batch_size'],
		metrics=[tm.Accuracy(), tm.F1Score()]
	)


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	hyperparameters = dict(
		hidden       = [10]
		, batch_size = [8]
		, epochs     = [5]
	)
	
	df = datum.to_df('epilepsy.parquet')
	label_df = df[['seizure']]
	label_dataset = Dataset.Tabular.from_df(label_df)
	
	seq_ndarray3D = df.drop(columns=['seizure']).to_numpy().reshape(1000,178,1)
	feature_dataset = Dataset.Sequence.from_numpy(seq_ndarray3D)

	pipeline = Pipeline(
		Input(
			dataset  = feature_dataset,
			encoders = Input.Encoder(sklearn_preprocess=StandardScaler())
		),
		
		Target(
			dataset = label_dataset
		),
		
		Stratifier(
			size_test       = 0.12, 
			size_validation = 0.22,
			fold_count      = fold_count
		)    
	)
	experiment = Experiment(
		Architecture(
			library           = "pytorch"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment



"""
High-Level API
├── Documentation = https://aiqc.readthedocs.io/en/latest/notebooks/api_high_level.html
└── Examples = https://aiqc.readthedocs.io/en/latest/tutorials.html

- The High-Level API is a declarative wrapper for the Low-Level API. 
  These few psuedo classes provide a logical way to group user input, 
  as opposed to tediously chaining many ORM objects together with relationships.
- It also allows us to make changes to the Low-Level API without 
  introducing major changes to the High-Level API.
- `__new__` is used in some places because `__init__` cannot return anything.
"""
from .orm import *
from .utils.wrangle import listify
#==================================================
# PIPELINE
#==================================================
class Target:
    def __init__(
        self, 
        dataset:object,
        column:str 			= None,
        interpolater:object = None,
        encoder:object		= None,
    ):
        """`column:str` in order to encourage single-column labels"""
        self.dataset = dataset
        self.column  = listify(column)

        l_id = Label.from_dataset(dataset_id=dataset.id, columns=column).id
        if (interpolater is not None):
            kwargz = interpolater.__dict__
            LabelInterpolater.from_label(label_id=l_id, **kwargz)
        if (encoder is not None): 
            kwargz = encoder.__dict__
            LabelCoder.from_label(label_id=l_id, **kwargz)
        self.id = l_id

    class Interpolater:
        def __init__(self, process_separately:bool=True, interpolate_kwargs:dict=None):
            self.process_separately = process_separately
            self.interpolate_kwargs = interpolate_kwargs
    
    class Encoder:
        def __init__(self, sklearn_preprocess:object):
            self.sklearn_preprocess = sklearn_preprocess


class Input:
    def __init__(
        self
        , dataset:object
        , exclude_columns:list 	= None
        , include_columns:list  = None
        , interpolaters:list	= None
        , window:object 		= None
        , encoders:list 		= None
        , reshape_indices:tuple = None
    ):
        self.dataset 		 = dataset
        self.exclude_columns = listify(exclude_columns)
        self.include_columns = listify(include_columns)
        self.interpolaters 	 = listify(interpolaters)
        self.window 		 = window
        self.encoders 		 = listify(encoders)
        self.reshape_indices = reshape_indices
    
    class Interpolater:
        def __init__(
            self
            , process_separately:bool = True
            , verbose:bool            = False
            , interpolate_kwargs:dict = None
            , dtypes:list             = None
            , columns:list            = None
        ):
            self.process_separately = process_separately
            self.verbose       		= verbose
            self.interpolate_kwargs = interpolate_kwargs
            self.dtypes        		= listify(dtypes)
            self.columns       		= listify(columns)
    
    class Window:
        def __init__(
            self
            , size_window:int
            , size_shift:int
            , record_shifted:bool = True # overwritten by pure inference.
        ):
            self.size_window    = size_window
            self.size_shift     = size_shift
            self.record_shifted = record_shifted
    
    class Encoder:
        def __init__(
            self
            , sklearn_preprocess:object
            , verbose:bool = False
            , include:bool = True
            , dtypes:list  = None
            , columns:list = None
        ):
            self.sklearn_preprocess = sklearn_preprocess
            self.verbose 			= verbose
            self.include 			= include
            self.dtypes 			= listify(dtypes)
            self.columns 			= listify(columns)


class Stratifier:
    def __init__(
        self
        , size_test:float 		= None
        , size_validation:float = None
        , fold_count:int 		= None
        , bin_count:int 		= None
    ):
        self.size_test 		 = size_test
        self.size_validation = size_validation
        self.fold_count 	 = fold_count
        self.bin_count 		 = bin_count


class Pipeline:
    def __new__(
        cls
        , inputs:list
        , target:object 	= None
        , stratifier:object = None
        , name:str			= None
        , description:str	= None
        , _predictor:int    = None
    ):					
        inputs = listify(inputs)

        if (target is not None):
            l_id      = target.id
            l_dset_id = target.dataset.id
        elif (target is None):
            # Need to know if label exists so it can be excluded.
            l_id      = None
            l_dset_id = None
        
        features = []
        feature_ids = []
        for i in inputs:
            d_id = i.dataset.id
           
            exclude_columns = i.exclude_columns
            include_columns = i.include_columns
            # For shared datasets, remove any label columns from featureset
            if (d_id==l_dset_id):
                l_cols = target.column
                
                if ((exclude_columns==None) and (include_columns==None)):
                    exclude_columns = l_cols
                # Both can't be set based on Feature logic
                elif (exclude_columns is not None):
                    for c in l_cols:
                        if (c not in exclude_columns):
                            exclude_columns.append(c)
                elif (include_columns is not None):
                    for c in l_cols:
                        if (c in include_columns):
                            include_columns.remove(c)

            f = Feature.from_dataset(dataset_id=d_id, exclude_columns=exclude_columns, include_columns=include_columns)
            f_id = f.id
            features.append(f)
            feature_ids.append(f.id)

            interpolaters = i.interpolaters
            if (interpolaters is not None):
                for fp in interpolaters:
                    kwargz = fp.__dict__
                    FeatureInterpolater.from_feature(f_id, **kwargz)

            # Window needs to be created prior to Splitset because it influences `samples`
            window = i.window
            if (window is not None):
                kwargz = window.__dict__
                Window.from_feature(feature_id=f_id, **kwargz)

            encoders = i.encoders
            if (encoders is not None):					
                for fc in encoders:
                    kwargz = fc.__dict__
                    FeatureCoder.from_feature(feature_id=f_id, **kwargz)
            
            reshape_indices = i.reshape_indices
            if (reshape_indices is not None):
                FeatureShaper.from_feature(feature_id=f_id, reshape_indices=reshape_indices)

        if (stratifier is None):
            # Initialize with Nones
            stratifier = Stratifier()
        
        if (_predictor is not None):
            _predictor = _predictor.id

        splitset = Splitset.make(
            feature_ids 	  = feature_ids
            , label_id 		  = l_id
            , size_test 	  = stratifier.size_test
            , size_validation = stratifier.size_validation
            , bin_count 	  = stratifier.bin_count
            , fold_count 	  = stratifier.fold_count
            , name 			  = name
            , description 	  = description
            , _predictor_id   = _predictor
        )
        return splitset

#==================================================
# EXPERIMENT
#==================================================
class Architecture:
    def __init__(
        self
        , library:str
        , analysis_type:str
        , fn_build:object
        , fn_train:object
        , fn_optimize:object   = None
        , fn_lose:object       = None
        , fn_predict:object    = None
        , hyperparameters:dict = None
    ):
        """Putting params here as `fn_*` can change when editing params"""
        self.hyperparameters = hyperparameters
        self.id = Algorithm.make(
            library 	  	= library
            , analysis_type = analysis_type
            , fn_build 	  	= fn_build
            , fn_train 	  	= fn_train
            , fn_optimize   = fn_optimize
            , fn_lose 	  	= fn_lose
            , fn_predict    = fn_predict
        ).id
        

class Trainer:
    def __init__(
        self
        , pipeline:object
        , repeat_count:int 	   = 1
        , permute_count:int    = 3
        , search_count:int 	   = None
        , search_percent:float = None
    ):
        """Intentionally switch to splitset here so it can be used in **kwargs"""
        self.splitset_id 	= pipeline.id
        self.repeat_count 	= repeat_count
        self.permute_count 	= permute_count
        self.search_count 	= search_count
        self.search_percent = search_percent


class Experiment:
    """
    - Create Algorithm, Hyperparamset, preprocess, and Queue.
    - Includes `preprocess` because it's weird to encode labels before you know what your final training layer looks like.
      Also, it's optional, so you'd have to access it from splitset before passing it in.
    - The only pre-existing things that need to be passed in are `splitset_id`.

    `encoder_feature`: List of dictionaries describing each encoder to run along with filters for different feature columns.
    `encoder_label`: Single instantiation of an sklearn encoder: e.g. `OneHotEncoder()` that gets applied to the full label array.
    """
    def __new__(cls, architecture:object, trainer:object):
        hyperparameters = architecture.hyperparameters
        if (hyperparameters is not None):
            h_id = Hyperparamset.from_algorithm(
                algorithm_id      = architecture.id
                , hyperparameters = hyperparameters
                , search_count    = trainer.search_count
                , search_percent  = trainer.search_percent
            ).id
        elif (hyperparameters is None):
            h_id = None

        kwargz = trainer.__dict__
        del kwargz['search_count'], kwargz['search_percent']
        queue = Queue.from_algorithm(
            algorithm_id = architecture.id
            , hyperparamset_id = h_id
            , **kwargz
        )
        return queue

#==================================================
# INFERENCE
#==================================================
class Inference:
    def __new__(
        cls
        , predictor:object
        , input_datasets:list
        , target_dataset:object = None # overwritten by during monitored inference
        , record_shifted:bool   = False # overwritten by during monitored inference
    ):
        """Reference `utils.wrangle.schemaNew_matches_schemaOld()` for validation"""
        input_datasets = listify(input_datasets)
        old_splitset = predictor.job.queue.splitset
        
        if (target_dataset is not None):
            label = old_splitset.label 
            if (label is not None):
                cols           = label.columns
                target_dataset = label.dataset
                target         = Target(dataset=target_dataset,column=cols)
        else:
            target = None

        inputs = []	
        for e, f in enumerate(old_splitset.features):		
            cols = f.columns
            dataset = input_datasets[e]
            new_window = None
            if (e==0):
                """
                - Window has to be created prior to splitset. 
                  So this can't happen during `Feature.preprocess()`
                - Reminder, unsupervised can only have 1 featureset right now.
                  So we only need to access the first Feature.
                """
                if (f.windows.count()>0):
                    old_window = f.windows[-1]

                    # Only pure inference is supported
                    new_window = Input.Window(
                        size_window      = old_window.size_window
                        , size_shift     = old_window.size_shift
                        , record_shifted = record_shifted
                    )
            # Use `include_columns` in case users decided to stop gathering the excluded columns.
            input_ = Input(
                dataset         = dataset, 
                include_columns = cols, 
                window          = new_window
            )
            inputs.append(input_)

        pipeline = Pipeline(
            inputs       = inputs
            , target     = target
            , stratifier = Stratifier()
            , _predictor = predictor
        )
        prediction = pipeline.infer()
        return prediction



"""PyTorch Binary Classification with Image data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..utils.config import app_folders, create_folder
from ..utils.ingest import imgFolder_to_arr4D
from ..utils.encoding import div255, mult255
from ..utils.pytorch import fit
from ..orm import Dataset
# External modules
from os import path
from numpy import save as np_save
import torch.nn as nn
import torchmetrics as tm
from sklearn.preprocessing import FunctionTransformer




def fn_build(features_shape, label_shape, **hp):
	#features_shape = (160, 120) after reshaping
	model = nn.Sequential(
		#Conv1d(in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros')
		nn.Conv1d(
			in_channels=features_shape[0]# is the width of the image
			, out_channels=56 #arbitrary number. treating this as network complexity.
			, kernel_size=3
			, padding=1
		)
		, nn.ReLU() #it wasnt learning with tanh
		, nn.MaxPool1d(kernel_size=2, stride=2)
		, nn.Dropout(p=0.4)

		, nn.Conv1d(
			in_channels=56, out_channels=128,
			kernel_size=3, padding=1
		)
		, nn.ReLU()
		, nn.MaxPool1d(kernel_size=2, stride=2)
		, nn.Dropout(p=0.4)
		#[5x3840]
		, nn.Flatten()
		, nn.Linear(3840,3840)
		, nn.BatchNorm1d(3840,3840)
		, nn.ReLU()
		, nn.Dropout(p=0.4)

		, nn.Linear(3840, label_shape[-1])
		, nn.Sigmoid()
	)
	return model


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):	return fit(
		model, loser, optimizer, 
		train_features, train_label,
		eval_features, eval_label,
		epochs=10, batch_size=5,
		metrics=[tm.Accuracy(),tm.F1Score()]
	)


def fn_predict(model, samples_predict):
	probability = model(samples_predict['features'])
	# Convert tensor back to numpy for AIQC metrics.
	probability = probability.detach().numpy()
	prediction = (probability > 0.5).astype("int32")
	# Both objects are numpy.
	return prediction, probability


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count=2):
	df = datum.to_df(name='brain_tumor.csv')
	label_dataset = Dataset.Tabular.from_df(dataframe=df)
	
	# Dataset.Image
	# Just ensuring we test all kinds of ingestion
	folder_path = 'remote_datum/image/brain_tumor/images'
	arr, _ = imgFolder_to_arr4D(folder_path)

	if (fold_count is None):
		feature_dataset = Dataset.Image.from_numpy(
			arr4D_or_npyPath=arr, retype='float64'
		)
		path_full = None
	else:
		path_models_cache = app_folders['cache_tests']
		create_folder(path_models_cache)
		path_file = f"temp_arr.npy"
		path_full = path.join(path_models_cache,path_file)
		np_save(path_full, arr, allow_pickle=True)
		feature_dataset = Dataset.Image.from_numpy(
			arr4D_or_npyPath=path_full, ingest=False, retype='float64'
		)


	pipeline = Pipeline(
		Input(
			dataset  = feature_dataset,
			encoders = Input.Encoder(FunctionTransformer(div255, inverse_func=mult255)),
			reshape_indices = (0,2,3)
		),

		Target(
			dataset = label_dataset,
			column  = 'status'
		),

		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)
	)
	
	experiment = Experiment(
		Architecture(
			library           = "pytorch"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = None
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment